from .controller import Upsonic_Remote
from .on_prem import Upsonic_On_Prem, Tiger, Tiger_Admin
from .interface import Upsonic_Cloud_Free
from .interface import Upsonic_Cloud_Pro
from .interface import Upsonic_Cloud_Premium
from .interface import Upsonic_Cloud_Startup
from .interface import Upsonic_Cloud_Readonly
from .interface import Upsonic_Cloud_Generic
from .helper import no_exception
from .helper import requires
from .interface import encrypt
from .interface import decrypt
from .interface import upsonic_serializer
from .localimport import localimport
from . import interface


import traceback

try:
    from .core import Upsonic
    from .core import start_location
    from .core import HASHES
    from .core import Upsonic_Serial

except:
    pass

from .remote import localimport
from .remote import Upsonic_Remote
from .remote import Upsonic_On_Prem, Tiger, Tiger_Admin
from .remote import Upsonic_Cloud_Free
from .remote import Upsonic_Cloud_Pro
from .remote import Upsonic_Cloud_Premium
from .remote import Upsonic_Cloud_Startup
from .remote import Upsonic_Cloud_Readonly
from .remote import Upsonic_Cloud_Generic
from .remote import no_exception
from .remote import requires
from .remote import encrypt
from .remote import decrypt
from .remote import upsonic_serializer
from .remote import interface


open_databases = {}

__version__ = '0.24.0'







#!/usr/bin/python3
# -*- coding: utf-8 -*-

import json
import ast
from functools import wraps
from hashlib import sha256


import pickle
import os
import re

import copy
from cryptography.fernet import Fernet
import base64
import hashlib
import inspect
import pkgutil
import threading
import time
import textwrap
import importlib.util
import cloudpickle

from contextlib import contextmanager

import sys

from rich.progress import Progress

import textwrap
import dill
import sys

from pip._internal.operations import freeze

import traceback
import os, hashlib, shutil
from memory_profiler import memory_usage



from rich.console import Console

console = Console()

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
from remote.localimport import localimport
from remote.interface import encrypt
from remote.interface import decrypt


def extract_needed_libraries(func, debug=False):
    result = {}
    the_globals = dill.detect.globalvars(func)
    for each in the_globals:
        name = dill.source.getname(the_globals[each], fqn=True)
        result[each] = name.split(".")[0]
    print("result", result) if debug else None
    return result


def extract_source(obj, debug=False):
    the_source = dill.source.findsource(obj)[0]
    print(the_source) if debug else None
    my_source = ""
    for each in the_source:
        my_source += each

    return my_source
def extract_local_files(obj, debug=False, local_directory=None):
    if local_directory == None:
        local_directory = os.getcwd()
    print(local_directory) if debug else None

    the_elements = dill.detect.globalvars(obj)
    print(the_elements) if debug else None
    the_local_elements = {}
    for element, value in the_elements.items():
        element_file = dill.source.getfile(value)
        print(element_file) if debug else None
        if element_file.startswith(local_directory):
            print("Inside") if debug else None
            with open(element_file, "r") as f:
                element_content = f.read()
            print("element_content", element_content) if debug else None
            the_local_elements[os.path.basename(element_file)] = element_content

    print("Complated") if debug else None
    return the_local_elements


def dump_local_files(extract, debug=False, local_directory=None):
    if local_directory == None:
        local_directory = os.getcwd()
    print(local_directory) if debug else None

    for element, value in extract.items():
        # Create a directory named as upsonic if not exists
        if not os.path.exists(os.path.join(local_directory, "upsonic")):
            os.makedirs(os.path.join(local_directory, "upsonic"))

        file_location = os.path.join(local_directory, "upsonic", element)
        print(file_location) if debug else None
        print(value) if debug else None
        with open(file_location, "w") as f:
            f.write(value)

        sys.path.insert(0, os.path.join(local_directory, "upsonic"))




class Upsonic_On_Prem:
    prevent_enable = False
    quiet_startup = False

    @staticmethod
    def export_requirement():
        the_list = list(freeze.freeze())
        the_string = ""
        for item in the_list:
            the_string += item + ", "
        return the_string[:-2]

    def _log(self, message):
        self.console.log(message)

    def __enter__(self):
        return self  # pragma: no cover

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass  # pragma: no cover


    @property
    def console(self):

        return console

    @property
    def localimport(self):

        return localimport

    @property
    def encrypt(self):

        return encrypt

    @property
    def decrypt(self):

        return decrypt


    def __init__(self, api_url, access_key, engine="cloudpickle,dill", enable_usage_analyses=True, enable_local_files=True, enable_elastic_dependency=False, cache_dir=None, pass_python_version_check=False, byref=True, recurse=True, protocol=pickle.DEFAULT_PROTOCOL, source=True, builtin=True, tester=False):
        import requests
        from requests.auth import HTTPBasicAuth

        from requests.packages.urllib3.exceptions import InsecureRequestWarning

        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)



        self.requests = requests
        self.HTTPBasicAuth = HTTPBasicAuth

        self.api_url = api_url
        self.password = access_key
        self.engine=engine
        self.byref=byref
        self.recurse=recurse
        self.protocol = protocol
        self.source = source
        self.builtin = builtin
        self.enable_elastic_dependency = enable_elastic_dependency
        self.enable_usage_analyses = enable_usage_analyses
        self.enable_local_files = enable_local_files

        self.tester = tester
        self.pass_python_version_check = pass_python_version_check

        self.enable_active = False

        self.cache_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "upsonic_cache") if cache_dir == None else cache_dir
        if not os.path.exists(self.cache_dir):
            os.mkdir(self.cache_dir)


        if self.status == True:
            self._log(
                f"[bold green]Upsonic[bold green] active",
            )
        else:
            self._log(
                f"[bold red]Upsonic[bold red] is down",
            )


        self.thread_number = 5

    def _send_request(self, method, endpoint, data=None, make_json=True):
        try:
            response = self.requests.request(
                method,
                self.api_url + endpoint,
                data=data,
                auth=self.HTTPBasicAuth("", self.password),
                verify=False,
            )
            try:
                result = None
                if not make_json:
                    result = response.text
                else:
                    result = json.loads(response.text)
                    if result["status"] == False:
                        self._log(
                            f"[bold red]Error: {endpoint}",
                        )
                    else:
                        result = result["result"]

                return result
            except:  # pragma: no cover
                print(f"Error on '{self.api_url + endpoint}': ", response.text) if self.tester else None
                return [None]  # pragma: no cover
        except:
            print("Error: Remote is down")
            return [None]

    @property
    def status(self):
        return self._send_request("GET", "/status")

    def get_specific_version(self, package):
        package_name = package.split("==")[0]
        package_version = (
            package.split("==")[1]
            if len(package.split("==")) > 1
            else "Latest"
        )
        backup_sys_path = sys.path
        backup_sys_modules = sys.modules

        the_dir = os.path.abspath(
            os.path.join(self.cache_dir, package_name, package_version)
        )
        with self.localimport(the_dir) as _importer:
            return importlib.import_module(package_name)

    def generate_the_globals(self, needed_libraries, key):

        requirements = self.extract_the_requirements(key)

        total = {}
        for each, value in needed_libraries.items():
            the_needed = None
            for each_r in requirements:
                each_r_ = each_r.split("==")[0]
                if each_r_.split(".")[0].lower() == value.split(".")[0].lower():
                    total[each] = self.get_specific_version(each_r.lower())

        return total

    def generate_the_true_requirements(self, requirements, needed_libraries, key):


        total = {}
        for each, value in needed_libraries.items():
            the_needed = None
            for each_r in requirements:
                each_r_ = each_r.split("==")[0]
                if each_r_.split(".")[0].lower() == value.split(".")[0].lower():
                    total[each] = each_r

        return total


    def install_package(self, package):
        from pip._internal import main as pip

        package_name = package.split("==")[0]
        package_version = (
            package.split("==")[1]
            if len(package.split("==")) > 1
            else "Latest"
        )

        the_dir = os.path.abspath(
            os.path.join(self.cache_dir, package_name, package_version)
        )
        if not os.path.exists(the_dir) or not self.enable_elastic_dependency:

            if self.enable_elastic_dependency:
                os.makedirs(the_dir)
                if self.tester:
                    self._log(f"Installing {package} to {the_dir}")
                pip(["install", package, "--target", the_dir, "--no-dependencies"])
            else:
                if self.tester:
                    self._log(f"Installing {package} to default_dir")
                pip(["install", package])

    def extract_the_requirements(self, key):
        the_requirements = self.get_requirements(key)
        elements = []
        for each in the_requirements.split(","):
            if "==" in each:
                the_requirement = textwrap.dedent(each)
                elements.append(the_requirement)
        return elements

    def install_the_requirements(self, the_requirements):
        installed_requirements = self.export_requirement()
        if self.tester:
            self._log(f"installed_requirements {installed_requirements}")
        for each in the_requirements:
            try:
                if each not in installed_requirements or self.enable_elastic_dependency:
                    self.install_package(each)
                else:
                    if self.tester:
                        self._log("Already installed in system")
            except:
                if self.tester:
                    self._log(f"Error on {each}")
                    traceback.print_exc()

    def delete_cache(self):
        shutil.rmtree(self.cache_dir)


    def set_the_library_specific_locations(self, the_requirements):

        the_all_dirs = []
        the_all_string = ""

        ordered_list = sorted(the_requirements)
        if self.tester:
            self._log(f"ordered_list {ordered_list}")

        for package in ordered_list:
            package_name = package.split("==")[0]
            package_version = (
                package.split("==")[1]
                if len(package.split("==")) > 1
                else "Latest"
            )
            the_all_string += package

            the_dir = os.path.abspath(
                os.path.join(self.cache_dir, package_name, package_version)
            )

            the_all_dirs.append(the_dir)
        if self.tester:
            self._log(f"the_all_string {the_all_string}")

        # Create folder with sha256 of the_all_string
        sha256_string = hashlib.sha256(the_all_string.encode('utf-8')).hexdigest()
        sha256_dir = os.path.join(self.cache_dir, sha256_string)
        already_exist = os.path.exists(sha256_dir)
        os.makedirs(sha256_dir, exist_ok=True)

        if not already_exist:
            # Copying all contents in the_all_dirs to sha256_dir
            for dir_path in the_all_dirs:
                for root, dirs, files in os.walk(dir_path):
                    for file in files:
                        # construct full file path
                        full_file_name = os.path.join(root, file)
                        # construct destination path
                        dest_file_name = sha256_dir + full_file_name[len(dir_path):]
                        # create directories if not present in destination
                        os.makedirs(os.path.dirname(dest_file_name), exist_ok=True)
                        # copy file
                        shutil.copy(full_file_name, dest_file_name)

        if self.tester:
            self._log(f"the sha256 of new directory {already_exist} {sha256_dir}")


        return sha256_dir



    def unset_the_library_specific_locations(self):
        sys.path = self.sys_path_backup


    @contextmanager
    def import_package(self, package):
        package_name = package.split("==")[0]
        package_version = (
            package.split("==")[1]
            if len(package.split("==")) > 1
            else "Latest"
        )

        the_dir = os.path.abspath(
            os.path.join(self.cache_dir, package_name, package_version)
        )

        if not os.path.exists(the_dir):
            self.install_package(package)

        sys_path_backup = sys.path.copy()

        sys.path.insert(0, the_dir)

        try:
            yield
        finally:
            sys.path = sys_path_backup

    def extend_global(self, name, value):
        globals()[name] = value


    def load_module(self, module_name, version=None):
        encryption_key = "u"

        version_check_pass = False
        the_all = self.get_all()
        original_name = module_name
        sub_module_name = False
        if "." in module_name:
            sub_module_name = module_name.replace(".", "_")
            module_name = sub_module_name

        the_all_imports = {}
        for i in the_all:
            original_i = i
            if "_upsonic_" in i:
                continue
            if sub_module_name != False:
                i = i.replace(original_name, module_name)
            name = i.split(".")
            if module_name == name[0]:
                try:
                    if not self.pass_python_version_check and not version_check_pass:
                        key_version = self.get_python_version(original_i)
                        currenly_version = self.get_currently_version()
                        if self.tester:
                            self._log(f"key_version {key_version}")
                            self._log(f"currenly_version {currenly_version}")
                        if key_version[0] == currenly_version[0] and key_version[0] == "3":
                            if self.tester:
                                self._log(f"Versions are same and 3")
                            if key_version[1] != currenly_version[1]:
                                        if self.tester:
                                            self._log("Minor versions are different")

                                        self._log(
                                            f"[bold orange]Warning: The versions are different, are you sure to continue")
                                        the_input = input("Yes or no (y/n)").lower()
                                        if the_input == "n":
                                            key_version = f"{key_version[0]}.{key_version[1]}"
                                            currenly_version = f"{currenly_version[0]}.{currenly_version[1]}"
                                            return "Python versions is different (Key == " + key_version + " This runtime == " + currenly_version + ")"
                                        if the_input == "y":
                                            version_check_pass = True
                except:
                    if self.tester:
                        traceback.print_exc()

                if version != None:
                    version_list_response = self.get_version_history(original_i)
                    version_list = []
                    for each_v in version_list_response:
                        version_list.append(each_v.replace(original_i+":", ""))


                    if version in version_list:
                        try:
                            the_all_imports[i] = self.get(
                                original_i,
                                version,
                                pass_python_version_control=True
                            )
                        except:
                            the_all_imports[i] = self.get(original_i, pass_python_version_control=True)
                else:
                    the_all_imports[i] = self.get(original_i, pass_python_version_control=True)

        import types

        def create_module_obj(dictionary):
            result = {}
            for key, value in dictionary.items():
                modules = key.split(".")
                current_dict = result
                for module in modules[:-1]:
                    if module not in current_dict:
                        current_dict[module] = types.ModuleType(module)
                    current_dict = vars(current_dict[module])
                current_dict[modules[-1]] = value

            return result

        generated_library = create_module_obj(the_all_imports)[module_name]

        return generated_library

    def dump_module(
            self,
            module_name,
            module,

    ):
        encryption_key = "u"
        top_module = module

        cloudpickle.register_pickle_by_value(top_module)

        sub_modules = []
        if hasattr(top_module, "__path__"):

            for importer, modname, ispkg in pkgutil.walk_packages(
                    path=top_module.__path__,
                    prefix=top_module.__name__ + ".",
                    onerror=lambda x: None,
            ):
                sub_modules.append(
                    importer.find_module(modname).load_module(modname)
                )
        else:
            sub_modules.append(top_module)

        threads = []

        the_list = []

        for sub_module in sub_modules:
            [
                the_list.append(obj)
                for name, obj in inspect.getmembers(sub_module)
            ]

        # Extract just functions and classes
        the_list = [
            i for i in the_list if inspect.isfunction(i) or inspect.isclass(i)
        ]
        # If the __module__ is not equal to module_name, remove it from the list

        the_list = [
            i for i in the_list if i.__module__.split(".")[0] == module_name
        ]

        my_list = []
        for element in copy.copy(the_list):
            if inspect.isfunction(element):
                name = element.__module__ + "." + element.__name__

            elif inspect.isclass(element):
                name = element.__module__ + "." + element.__name__
            if (
                    not "upsonic.remote" in name
                    and not "upsonic_updater" in name
                    and name != f"{module.__name__}.threading.Thread"
            ):
                my_list.append(element)

        the_list = my_list

        with Progress() as progress:

            task1 = progress.add_task(
                "           [red]Job Started...", total=len(the_list)
            )
            task2 = progress.add_task(
                "           [green]Job Complated...", total=len(the_list)
            )

            for element in the_list:
                time.sleep(0.1)
                if inspect.isfunction(element):
                    name = element.__module__ + "." + element.__name__

                elif inspect.isclass(element):
                    name = element.__module__ + "." + element.__name__
                else:
                    continue

                first_element = name.split(".")[0]

                if first_element != module_name:
                    continue

                try:
                    while len(threads) >= self.thread_number:
                        for each in threads:
                            if not each.is_alive():
                                threads.remove(each)
                        time.sleep(0.1)

                    the_thread = threading.Thread(
                        target=self.set,
                        args=(name, element),
                    )
                    the_thread.start()

                    thread = the_thread
                    threads.append(thread)
                    progress.update(task1, advance=1)

                except:
                    import traceback

                    traceback.print_exc()
                    self._log(f"[bold red]Error on '{name}'")
                    self.delete(name)

            for each in threads:
                progress.update(task2, advance=1)
                each.join()

    def dump(
            self,
            key,
            value,

    ):

        return self.set(
            key,
            value,
        )

    def load(
            self,
            key,
            version=None
    ):
        return self.get(
            key,
            version=version,
            print_exc=True
        )


    def get_currently_version(self):
        total = sys.version_info
        the_version = []
        the_version.append(total.major)
        the_version.append(total.minor)
        the_version.append(total.micro)
        return the_version

    def get_python_version(self, key):
        data = {"scope": key}
        total = self._send_request("POST", "/get_python_version_of_scope", data)
        the_version = []
        the_version.append(int(total.split(".")[0]))
        the_version.append(int(total.split(".")[1]))
        the_version.append(int(total.split(".")[2]))
        return the_version


    def set(
            self,
            key,
            value,
    ):

        if key.startswith("."):
            self._log("Error: The key can not start with '.'")
            return False
        if ":" in key:
            self._log("Error: The key can not include ':'")
            return False
        if key.endswith("."):
            self._log("Error: The key can not end with '.'")
            return False
        if "." not in key:
            self._log("Error: You should create a parent with '.' like math.sum")
            return False
        if ".." in key:
            self._log("Error: The key can not include multiple dot one after one'..'")
            return False





        the_type = type(value).__name__
        if the_type == "type":
            the_type = "class"

        encryption_key = "u"

        data = {
            "scope": key,
            "code": textwrap.dedent(self.extract_source(value)),
        }

        self._send_request("POST", "/dump_code", data)





        data = {"scope": key, "type": the_type}

        self._send_request("POST", "/dump_type", data)


        the_requirements = Upsonic_On_Prem.export_requirement()
        the_original_requirements = ""
        if self.tester:
            self._log(f"The first original requirements {the_original_requirements}")
        elements = []
        for each in the_requirements.split(","):
            if "==" in each:
                the_requirement = textwrap.dedent(each)
                elements.append(the_requirement)
        the_requirements = elements
        if self.tester:
            self._log(f"the_requirements {the_requirements}")

        extracted_needed_libraries = None
        try:
            extracted_needed_libraries = extract_needed_libraries(value, self.tester)
            try:
                the_original_requirements = self.generate_the_true_requirements(the_requirements, extracted_needed_libraries, key)
                if self.tester:
                    self._log(f"the_original_requirements in_generation {the_original_requirements}")
                the_text = ""
                for each, value_ in the_original_requirements.items():
                    the_text += value_ + ", "
                the_original_requirements = the_text[:-2]

            except:
                if self.tester:
                    self._log(f"Error on generate_the_true_requirements while dumping {key}")
                    traceback.print_exc()
        except:
            if self.tester:
                self._log(f"Error on extract_needed_libraries while dumping {key}")
                traceback.print_exc()



        if self.tester:
            self._log(f"the_original_requirements {the_original_requirements}")
        data = {
            "scope": key,
            "requirements": the_original_requirements,
        }

        self._send_request("POST", "/dump_requirements", data)



        data = {
            "scope": key,
            "python_version": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        }

        self._send_request("POST", "/dump_python_version", data)

        fernet_key = base64.urlsafe_b64encode(hashlib.sha256(encryption_key.encode()).digest())
        fernet = Fernet(fernet_key)


        engine_reports_exceptions = []
        the_engine_reports = {}
        for engine in self.engine.split(","):
            try:
                the_engine_reports[engine] = self.encrypt(encryption_key, value, engine, self.byref, self.recurse, self.protocol, self.source, self.builtin)
            except:
                if self.tester:
                    self._log(f"Error on {engine} while dumping {key}")
                    traceback.print_exc()
                else:
                    engine_reports_exceptions.append(traceback.format_exc())


        if len(the_engine_reports) == 0 and not self.tester:
            self._log("[red] Error: No engine is able to dump the object")
            for error in engine_reports_exceptions:
                print(error)

        try:
            the_engine_reports["extracted_local_files"] = fernet.encrypt(pickle.dumps(extract_local_files(value, self.tester), protocol=1))
        except:
            if self.tester:
                self._log(f"Error on extracted_local_files while dumping {key}")
                traceback.print_exc()

        try:
            the_engine_reports["extract_source"] = fernet.encrypt(pickle.dumps(extract_source(value, self.tester), protocol=1))
        except:
            if self.tester:
                self._log(f"Error on extract_source while dumping {key}")
                traceback.print_exc()

        try:
            the_engine_reports["extract_source"] = fernet.encrypt(pickle.dumps(extract_source(value, self.tester), protocol=1))
        except:
            if self.tester:
                self._log(f"Error on extract_source while dumping {key}")
                traceback.print_exc()

        if extracted_needed_libraries != None:
            the_engine_reports["extract_needed_libraries"] = fernet.encrypt(pickle.dumps(extracted_needed_libraries, protocol=1))


        if self.tester:
            self._log(f"the_engine_reports {the_engine_reports}")
        dumped = pickle.dumps(the_engine_reports, protocol=1)


        data = {
            "scope": key,
            "data": fernet.encrypt(dumped)
        }

        response = self._send_request("POST", "/dump", data)

        if response != [None]:
            return True
        else:
            return False

    def get(
            self,
            key,
            version=None,
            print_exc=True,
            pass_python_version_control=False,
            pass_usage_analyses=False,
            try_to_extract_importable=False

    ):
        if self.tester:
            self._log(f"Process started for {key}")
        response = None

        encryption_key = "u"

        data = {"scope": key}

        versions_are_different = False
        if pass_python_version_control:
            versions_are_different = True
        try:
            if not self.pass_python_version_check and not pass_python_version_control:
                key_version = self.get_python_version(key)
                currenly_version = self.get_currently_version()
                if self.tester:
                    self._log(f"key_version {key_version}")
                    self._log(f"currenly_version {currenly_version}")
                if key_version[0] == currenly_version[0] and key_version[0] == 3:
                    if self.tester:
                        self._log(f"Versions are same and 3")
                    if key_version[1] != currenly_version[1]:
                        if self.tester:
                            self._log("Minor versions are different")
                        if int(currenly_version[1]) >= 11 or int(key_version[1]) >= 11:
                            if int(currenly_version[1]) < 11 or int(key_version[1]) < 11:
                                versions_are_different = True
                                self._log(f"[bold orange]Warning: The versions are different, are you sure to continue")
                                the_input = input("Yes or no (y/n)").lower()
                                if the_input == "n":
                                    key_version = f"{key_version[0]}.{key_version[1]}"
                                    currenly_version = f"{currenly_version[0]}.{currenly_version[1]}"
                                    return "Python versions is different (Key == " + key_version + " This runtime == " + currenly_version + ")"
        except:
            if self.tester:
                traceback.print_exc()
        the_requirements_path = None

        try:
                the_requirements = self.extract_the_requirements(key)

                self.install_the_requirements(the_requirements)
                if self.tester:
                    self._log(f"the_requirements {the_requirements}")
                if self.enable_elastic_dependency:
                    the_requirements_path = self.set_the_library_specific_locations(the_requirements)
        except:
                if self.tester:
                    self._log(f"Error on requirements while dumping {key}")
                    traceback.print_exc()

        if response is None:
            if version != None:
                response = self.get_version_data(key, version)
            else:
                response = self._send_request("POST", "/load", data)
        try:
            fernet_key = base64.urlsafe_b64encode(hashlib.sha256(encryption_key.encode()).digest())
            fernet = Fernet(fernet_key)
            response = pickle.loads(fernet.decrypt(response))
            if self.tester:
                self._log(f"response {response}")
            if "extracted_local_files" in response:
                try:
                    if self.enable_local_files:
                        dump_local_files(pickle.loads(fernet.decrypt(response["extracted_local_files"])), self.tester)
                    else:
                        if self.tester:
                            self._log(f"Local files are not enabled")
                except:
                    if self.tester:
                        self._log(f"Error on extracted_local_files while loading {key}")
                        traceback.print_exc()
                response.pop("extracted_local_files")
            if "extract_source" in response:
                response.pop("extract_source")
            needed_libraries = None
            if "extract_needed_libraries" in response:
                needed_libraries = pickle.loads(fernet.decrypt(response["extract_needed_libraries"]))
                response.pop("extract_needed_libraries")
            for engine, value in response.items():

                try:
                    if the_requirements_path is not None:
                        with self.localimport(the_requirements_path) as _importer:
                            response = self.decrypt(encryption_key, value, engine, try_to_extract_importable=try_to_extract_importable)
                            break
                    else:
                        response = self.decrypt(encryption_key, value, engine, try_to_extract_importable=try_to_extract_importable)
                        break
                except:
                    response = "Error"
                    self._log(f"Error on {engine} while loading {key}")
                    traceback.print_exc()
        except:
            if print_exc:
                if self.tester:
                    traceback.print_exc()
            else:
                pass


        #Run anayses
        if self.enable_usage_analyses and not pass_usage_analyses:
            if inspect.isfunction(response) and self.is_usage_analyses_true(key):
                commit = None
                if version == None:
                    commit = self.get_dump_history(key)[0].split(":")[1]
                response = self.profile_function(key, version, commit, response)


        return response

    def add_run_history(self, key, version, cpu_usage_one_core, memory_usage, elapsed_time, type):
        data = {
            "scope": key,
            "version": version,
            "cpu_usage": cpu_usage_one_core,
            "memory_usage": memory_usage,
            "elapsed_time": elapsed_time,
            "type": type
        }

        self._send_request("POST", "/dump_run", data)


    def get_dump_history(self, key):
        data = {"scope": key}
        return self._send_request("POST", "/get_dump_history", data)

    def profile_function(self, key, version, commit, func):
        @wraps(func)
        def wrapper_function(*args, **kwargs):
            # Get current process time and memory usage before function execution
            start_time = time.process_time()
            start_memory = memory_usage(-1, interval=0.1, timeout=1)[0]
            # Run the function and get its output
            succed = True
            try:
                output = func(*args, **kwargs)
            except:
                traceback.print_exc()
                succed=False
                output = traceback.format_exc()
            # Get process time and memory usage after function execution
            end_time = time.process_time()
            end_memory = memory_usage(-1, interval=0.1, timeout=1)[0]
            # Calculate time used and memory used
            time_used = end_time - start_time
            memory_used = end_memory - start_memory
            # Calculate the total CPU time available
            total_time = time_used
            # Calculate the CPU usage of the function
            total_cpu_count = os.cpu_count()
            cpu_usage_for_one_core = time_used * 100
            cpu_usage_total = cpu_usage_for_one_core / os.cpu_count()
            # print CPU and memory usage
            #print(f'CPU used by one core: {cpu_usage_for_one_core}%')
            #print(f'CPU used by total cores ({total_cpu_count}): {cpu_usage_total}%')
            #print(f'Memory used: {memory_used} MiB')
            #print(f'Time consuming by second: {total_time}')
            # Return the function output
            the_version = None
            if version == None:
                latest_commit = commit
                the_version = latest_commit
            else:
                the_version = version
            the_type = "Succed" if succed else "Failed"
            try:
                self.add_run_history(key, the_version, cpu_usage_for_one_core, memory_used, total_time, the_type)
            except:
                self._log(f"Error on adding run history, for server not supported. {key}")
            return output

        return wrapper_function

    def get_settings(self, key):
        data = {"scope": key}
        return self._send_request("POST", "/get_settings_of_scope", data)

    def is_usage_analyses_true(self, key):
        settings = self.get_settings(key)

        if settings == None or settings == [None]:
            cpu_usage_analyses = False
        else:
            if "usage_analyses" in settings:
                try:
                    cpu_usage_analyses = settings["usage_analyses"].lower() == "true"
                except:

                    cpu_usage_analyses = None

        return cpu_usage_analyses


    def active(
            self,
            value=None,

    ):
        encryption_key = "u"

        def decorate(value):
            key = value.__name__
            if (
                    value.__module__ != "__main__"
                    and value.__module__ != None
                    and not just_name
            ):
                key = value.__module__ + "." + key
            self.set(
                key,
                value,
            )

        if value == None:
            return decorate
        else:
            decorate(value)
            return value

    def get_all(self, ):
        encryption_key = "u"

        datas = self._send_request("GET", "/get_all_scopes_user")
        return datas

    def delete(self, scope):
        data = {"scope": scope}
        return self._send_request("POST", "/delete_scope", data)

    def database_list(self):
        return ast.literal_eval(self._send_request("GET", "/database/list"))

    def database_rename(self, database_name, new_database_name):
        data = {
            "database_name": database_name,
            "new_database_name": new_database_name,
        }
        return self._send_request("POST", "/database/rename", data)

    def database_pop(self, database_name):
        data = {"database_name": database_name}
        return self._send_request("POST", "/database/pop", data)

    def database_pop_all(self):
        return self._send_request("GET", "/database/pop_all")

    def database_delete(self, database_name):
        data = {"database_name": database_name}
        return self._send_request("POST", "/database/delete", data)

    def database_delete_all(self):
        return self._send_request("GET", "/database/delete_all")


    def ai_completion(self, message, model=None):
        data = {"message": message}
        if model != None:
            data["model"] = model
        return self._send_request("POST", "/ai_completion", data)

    def get_all_scopes_user(self):
        return self._send_request("GET", "/get_all_scopes_user")


    def extract_source(self, value):
        result = ""
        try:
            result = inspect.getsource(value)
        except:
            result = dill.source.getsource(value)
        return result


    def auto_dump(self, value, ask=True, check_idea=True, print_prompts=False, model=None):
        if check_idea:
            check = self.check_idea(value, print_prompts=print_prompts, model=model)
            if check != True:
                print("Check:", check)
                return

        code = textwrap.dedent(self.extract_source(value))
        all_scopes = self.get_all_scopes_user()
        all_scopes = "\n".join(all_scopes)

        prompt = f"""
You are an helpful software engineer. Help to organize library elements in a short and clear manner.


Generate a position for this:
```
{code}
```

Currenlty Index of Library:
```
{all_scopes}
```

Suggested Position:

"""

        ai_answer = self.ai_completion(prompt, model=model)
        ai_answer = ai_answer.replace("`", "").replace("\n", "")
        ai_answer = '.'.join(ai_answer.split('.')[:-1])
        ai_answer = ai_answer + "." + dill.source.getname(value)
        prompt = prompt + f"\nASSISTANT: {ai_answer}\n"

        prompt = prompt + f"\nQUESTION: Extract and just answer with the suggested position"
        if print_prompts:
            print("Prompt", prompt.replace(code, "CODE").replace(all_scopes, "ALL SCOPES"))
        ai_answer = self.ai_completion(prompt, model=model)
        if print_prompts:
            print("AI answer", ai_answer.replace(code, "CODE").replace(all_scopes, "ALL SCOPES"))
        ai_answer = ai_answer.replace("`", "").replace("\n", "")
        ai_answer = ai_answer.replace("ASSISTANT: ", "")
        if ai_answer in all_scopes:
            print(f"Check: similarity with the {ai_answer} is detected")
            return
        if ask:
            print("Commands:\n(Y)es/(N)o\n")
            while True:
                y_n = input(f"{ai_answer} ").lower()

                if y_n == "y":
                    self.set(ai_answer, value)
                    print("\nDumped")
                    break
                if y_n == "n":
                    break

        else:
            self.set(ai_answer, value)
            print("\nDumped")


    def get_code(self, scope):
        data = {"scope": scope}
        return self._send_request("POST", "/get_code_of_scope", data)

    def get_document(self, scope, version=None):
        data = {"scope": scope}
        if version != None:
            data["version"] = version
        return self._send_request("POST", "/get_document_of_scope", data)


    def check_idea(self, value, print_prompts=False, model=None):
        code = textwrap.dedent(self.extract_source(value))


        all_scopes_ = self.get_all_scopes_user()
        all_scopes = ""
        for i in all_scopes_:
            all_scopes += i + "\n"

        if print_prompts:
            print("Code", code)
            print("All scopes", all_scopes)

        prompt = f"""
Current Library Index:
```
{all_scopes}

```

Now analyze the each element of Current Library Index, if you want a potential similar functionality with this:

```
{code}

```

Which one is the most similar ?
"""

        ai_answer = self.ai_completion(prompt, model=model)

        ai_answer = ai_answer.replace("`", "").replace("\n", "")


        similarity_explanation = ai_answer

        prompt = prompt + f"\nASSISTANT: {ai_answer}\n"

        prompt = prompt + f"\nQUESTION: Is there any duplication risk (Y/N)?"


        if print_prompts:
            print("Prompt", prompt.replace(code, "CODE").replace(all_scopes, "ALL SCOPES"))
        ai_answer = self.ai_completion(prompt, model=model)
        ai_answer = ai_answer.replace("`", "").replace("\n", "")
        ai_answer = ai_answer.split(",")[0]
        ai_answer = ai_answer.replace("ASSISTANT: ", "")
        if print_prompts:
            print("AI answer", ai_answer.replace(code, "CODE").replace(all_scopes, "ALL SCOPES"))
        if ai_answer == "Y" or ai_answer == "YES" or ai_answer == "Yes":
            prompt = prompt + f"\nASSISTANT: {ai_answer}\n"

            prompt = prompt + f"\nQUESTION: Extract and just answer with the suggested position"
            ai_answer = self.ai_completion(prompt, model=model)
            ai_answer = ai_answer.replace("`", "").replace("\n", "")
            ai_answer = ai_answer.split(",")[0]
            ai_answer = ai_answer.replace("ASSISTANT: ", "")
            return "similarity: "+ai_answer + " - " + similarity_explanation
        if ai_answer == "N" or ai_answer == "NO" or ai_answer == "No":
            return True
        return similarity_explanation

    def search_by_documentation(self, question):
        data = {"question": question}
        response = self._send_request("POST", "/search_by_documentation", data)
        result = []
        for i in response:
            result.append(i[0])
        return result

    def search(self, question):
        return self.search_by_documentation(question)


    def get_default_ai_model(self):
        return self._send_request("GET", "/get_default_ai_model")


    def get_version_history(self, key):
        data = {"scope": key}
        version = self._send_request("POST", "/get_version_history", data)
        return version

    def get_module_version_history(self, key):
        data = {"top_library": key}
        return self._send_request("POST", "/get_module_version_history", data)


    def delete_version(self, key, version):
        data = {"version": key+":"+version}
        return self._send_request("POST", "/delete_version", data)

    def delete_module_version(self, module_name, version):
        data = {"top_library":module_name ,"version": version}
        return self._send_request("POST", "/delete_version_prefix", data)

    def create_version(self, key, version):
        data = {"scope":key ,"version": version}
        return self._send_request("POST", "/create_version", data)


    def create_module_version(self, module_name, version):
        data = {"top_library":module_name ,"version": version}
        return self._send_request("POST", "/create_version_prefix", data)




    def get_version_data(self, key, version):
        data = {"version": key+":"+version}
        return self._send_request("POST", "/load_specific_version", data)
    
    
    def get_requirements(self, key):
        data = {"scope": key}
        return self._send_request("POST", "/get_requirements_of_scope", data)

    def add_requirement(self, key, requirement):
        currently_requirements = self.get_requirements(key)
        if currently_requirements == None:
            currently_requirements = ""

        if currently_requirements != "":
            currently_requirements += ", "

        currently_requirements += requirement
        data = {
            "scope": key,
            "requirements": currently_requirements,
        }
        return self._send_request("POST", "/dump_requirements", data)

    def clear_requirements(self, key):
        data = {
            "scope": key,
            "requirements": "",
        }
        return self._send_request("POST", "/dump_requirements", data)

    def get_type(self, key, version=None):
        data = {"scope": key}
        if version != None:
            data["version"] = version
        return self._send_request("POST", "/get_type_of_scope", data)

    def get_code(self, key, version=None):
        data = {"scope": key}
        if version != None:
            data["version"] = version
        return self._send_request("POST", "/get_code_of_scope", data)


    def langchain(self, prefix=None, version=None):
        from langchain.pydantic_v1 import BaseModel, Field
        from langchain.tools import BaseTool, StructuredTool, tool
        all_functions = []
        for each in self.get_all():
            if each.endswith("__user"):
                continue
            the_true_name = each
            the_true_name = the_true_name.replace("_", ".")
            the_true_name = the_true_name.replace(".", "_")
            if prefix != None:
                if not each.startswith(prefix):
                    continue
            if self.get_type(each) == "function":
                the_function = self.get(each, version=version)
                if inspect.isfunction(the_function):
                    the_document = self.get_document(each, version=version) or " "
                    the_function.__doc__ = the_document
                    try:
                        the_tool = tool(the_function)
                        original_name = the_tool.name
                        the_tool.name = the_true_name
                        the_tool.description = the_tool.description.replace(original_name, the_true_name)[:1000]
                        all_functions.append(the_tool)
                    except:
                        traceback.print_exc()
                        pass
        return all_functions



    def crewai(self, prefix=None, version=None):
        from crewai_tools import tool
        all_functions = []
        for each in self.get_all():
            if each.endswith("__user"):
                continue
            the_true_name = each
            the_true_name = the_true_name.replace("_", ".")
            the_true_name = the_true_name.replace(".", "_")
            if prefix != None:
                if not each.startswith(prefix):
                    continue
            if self.get_type(each) == "function":
                the_function = self.get(each, version=version)
                if inspect.isfunction(the_function):
                    the_document = self.get_document(each, version=version) or " "
                    the_function.__doc__ = the_document
                    try:
                        the_tool = tool(the_function)
                        original_name = the_tool.name
                        the_tool.name = the_true_name
                        the_tool.description = the_tool.description.replace(original_name, the_true_name)[:1000]
                        all_functions.append(the_tool)
                    except:
                        traceback.print_exc()
                        pass
        return all_functions


    def autogen(self, caller, executor, prefix=None, version=None):
        import autogen
        for each in self.get_all():
            if each.endswith("__user"):
                continue
            the_true_name = each
            the_true_name = the_true_name.replace("_", ".")
            the_true_name = the_true_name.replace(".", "_")
            if prefix != None:
                if not each.startswith(prefix):
                    continue
            if self.get_type(each) == "function":
                the_function = self.get(each, version=version)
                if inspect.isfunction(the_function):
                    the_document = self.get_document(each, version=version) or " "
                    the_function.__doc__ = the_document
                    try:
                        the_document = the_document.replace(each, the_true_name).replace(each.split(".")[-1], the_true_name)[:1000]
                        autogen.agentchat.register_function(
                            the_function,
                            caller=caller,
                            executor=executor,
                            description=the_document,
                            name=the_true_name
                        )
                    except:
                        traceback.print_exc()
                        pass


    def openinterpreter(self, agent, prefix=None, version=None):
        def replace_function_name(input_string, new_function_name):
            result = re.sub(r"^(def )(\w+)", f"\g<1>{new_function_name}", input_string, flags=re.MULTILINE)
            return result

        def extract_function_definition(input_string):
            function_name = input_string.split('def ')[1].split(')')[0]

            return function_name + ")"

        class open_interpreter_tool:
            def __init__(self, function, name, document):
                self.name = name
                self.description = document
                self.function = replace_function_name(function, name)

            @property
            def type_for_prompt(self):
                the_description = self.description[:200]
                the_description.replace("\n", " ")
                return extract_function_definition(self.function).replace("\n", " ") + " #" + the_description

        all_functions = []
        for each in self.get_all():
            if each.endswith("__user"):
                continue
            the_true_name = each
            original_name = each
            the_true_name = the_true_name.replace("_", ".")
            the_true_name = the_true_name.replace(".", "_")
            if prefix != None:
                if not each.startswith(prefix):
                    continue
            if self.get_type(each) == "function":
                the_function = self.get(each, version=version, try_to_extract_importable=True)
                the_document = self.get_document(each, version=version) or " "
                the_document = the_document.replace(original_name.split(".")[-1], the_true_name)[:1000]
                try:
                    the_tool = open_interpreter_tool(the_function, the_true_name, the_document)
                    all_functions.append(the_tool)
                except:
                    traceback.print_exc()
                    pass

        for each_f in all_functions:
            if self.tester:
                self._log(each_f.function)
            agent.computer.run("python", each_f.function)

        agent.system_message += r"""
# THE OTHER APIs
    
These functions ALREADY IMPORTED, and can be used for many tasks:
    
```python
    
    """

        for each_f in all_functions:
            agent.system_message += each_f.type_for_prompt + "\n\n"

        agent.system_message += r"""
```
    
Do not import the anythink, They are already imported.
    """



def Tiger():
    return Upsonic_On_Prem("https://api_tiger.upsonic.co", 'ACK_xmxIiqsgGySvBPPd55M0Ldm5AcR2kt6r3kmL52Ptqo', engine="upsonic_serializer", pass_python_version_check=True)


def Tiger_Admin(api_url, access_key):
    return Upsonic_On_Prem(api_url, access_key, engine="upsonic_serializer", pass_python_version_check=True)


#!/usr/bin/python3
# -*- coding: utf-8 -*-


from dotenv import load_dotenv
load_dotenv(dotenv_path=".env")
import os
import inspect

import textwrap

import cloudpickle
import dill
import pickle
import importlib.util


def upsonic_serializer(func):
    the_source = dill.source.findsource(func)
    the_full_string = ""
    for each in the_source[0]:
        the_full_string += each
    imports = [line + "\n" for line in the_full_string.split('\n')
               if line.startswith('import ') or line.startswith('from ')]

    the_import_string = ""
    for each in imports:
        the_import_string += each

    the_function_string = dill.source.getsource(func)

    return the_import_string + "\n" + the_function_string

def encrypt(key, message, engine, byref, recurse, protocol, source, builtin):



    from cryptography.fernet import Fernet
    import base64
    import hashlib
    
    fernet_key = base64.urlsafe_b64encode(hashlib.sha256(key.encode()).digest())
    fernet = Fernet(fernet_key)

    dumped = None
    if engine == "cloudpickle":
        the_module = dill.detect.getmodule(message)
        if the_module is not None:
            cloudpickle.register_pickle_by_value(the_module)
        dumped = cloudpickle.dumps(message, protocol=protocol)
    elif engine == "dill":
        dumped = dill.dumps(message, protocol=protocol, byref=byref, recurse=recurse)
    elif engine == "upsonic_serializer":

        name_of_object = dill.source.getname(message)

        if name_of_object == None:
            try:
                name_of_object = message.__name__
            except:
                pass

        dumped = {"name": name_of_object, "upsonic_serializer": upsonic_serializer(message)}
        dumped = pickle.dumps(dumped, protocol=1)


    encrypted_message = fernet.encrypt(dumped)
    return encrypted_message

def decrypt(key, message, engine, try_to_extract_importable=False):
    from cryptography.fernet import Fernet
    import base64
    import hashlib    
    fernet = Fernet(base64.urlsafe_b64encode(hashlib.sha256(key.encode()).digest()))

    loaded = None
    if engine == "cloudpickle":
        loaded = cloudpickle.loads(fernet.decrypt(message))
    elif engine == "dill":
        loaded = dill.loads(fernet.decrypt(message))
    elif engine == "upsonic_serializer":

        loaded = pickle.loads(fernet.decrypt(message))

        if try_to_extract_importable:
            return loaded["upsonic_serializer"]

        def extract(code_string, function_name):
            tmp_dir = os.path.dirname(os.path.abspath(__file__))
            tmp_file = os.path.join(tmp_dir, function_name+"_upsonic" + ".py")
            with open(tmp_file, "w") as f:
                f.write(code_string)

            spec = importlib.util.spec_from_file_location(function_name+"_upsonic", tmp_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            os.remove(tmp_file)  # Clean up the temporary file

            return getattr(module, function_name)

        loaded = extract(loaded["upsonic_serializer"], loaded["name"])

    return loaded



def Upsonic_Cloud_Free(database_name=None, access_key=None, locking=None, client_id=None, cache=None, cache_counter=5, version=None, client_version=None, key_encyption=None, meta_datas=None, quiet=False, thread_number=None):
    if database_name == None:
        database_name = os.environ.get("database_key")
    if access_key == None:
        access_key = os.environ.get("access_key")


    if locking == None:
        locking = os.environ.get("locking", "false").lower() == "true"


    if client_id == None:
        client_id = os.environ.get("client_id", None)

    if cache == None:
        cache = os.environ.get("cache", "true").lower() == "true"
    if cache_counter == None:
        cache_counter = int(os.environ.get("cache_counter", "5"))



    if version == None:
        version = os.environ.get("version", "false").lower() == "true"

    if client_version == None:
        client_version = os.environ.get("client_version", "false").lower() == "true"


    if key_encyption == None:
        key_encyption = os.environ.get("key_encyption", "false").lower() == "true"



    if meta_datas == None:
        meta_datas = os.environ.get("meta_datas", "true").lower() == "true"


    if thread_number == None:
        thread_number = int(os.environ.get("thread_number", "1"))


    from upsonic import Upsonic_Remote
    return Upsonic_Remote(
        database_name, "https://cloud_1.upsonic.co", access_key, verify=True, locking=locking, client_id=client_id, cache=cache, cache_counter=cache_counter, version=version, client_version=client_version, key_encyption=key_encyption, meta_datas=meta_datas, quiet=quiet, thread_number=thread_number
    )  # pragma: no cover


def Upsonic_Cloud_Pro(database_name=None, access_key=None, locking=None, client_id=None, cache=None, cache_counter=5, version=None, client_version=None, key_encyption=None, meta_datas=None, quiet=False, thread_number=None):
    if database_name == None:
        database_name = os.environ.get("database_key")
    if access_key == None:
        access_key = os.environ.get("access_key")


    if locking == None:
        locking = os.environ.get("locking", "false").lower() == "true"


    if client_id == None:
        client_id = os.environ.get("client_id", None)

    if cache == None:
        cache = os.environ.get("cache", "true").lower() == "true"
    if cache_counter == None:
        cache_counter = int(os.environ.get("cache_counter", "5"))


    if version == None:
        version = os.environ.get("version", "false").lower() == "true"

    if client_version == None:
        client_version = os.environ.get("client_version", "false").lower() == "true"

    if key_encyption == None:
        key_encyption = os.environ.get("key_encyption", "false").lower() == "true"


    if meta_datas == None:
        meta_datas = os.environ.get("meta_datas", "true").lower() == "true"

    if thread_number == None:
        thread_number = int(os.environ.get("thread_number", "1"))

    from upsonic import Upsonic_Remote
    return Upsonic_Remote(
        database_name, "https://cloud_2.upsonic.co", access_key, verify=True, locking=locking, client_id=client_id, cache=cache, cache_counter=cache_counter, version=version, client_version=client_version, key_encyption=key_encyption, meta_datas=meta_datas, quiet=quiet, thread_number=thread_number
    )  # pragma: no cover


def Upsonic_Cloud_Premium(database_name=None, access_key=None, locking=None, client_id=None, cache=None, cache_counter=5, version=None, client_version=None, key_encyption=None, meta_datas=None, quiet=False, thread_number=None):
    if database_name == None:
        database_name = os.environ.get("database_key")
    if access_key == None:
        access_key = os.environ.get("access_key")

    if locking == None:
        locking = os.environ.get("locking", "false").lower() == "true"


    if client_id == None:
        client_id = os.environ.get("client_id", None)

    if cache == None:
        cache = os.environ.get("cache", "true").lower() == "true"
    if cache_counter == None:
        cache_counter = int(os.environ.get("cache_counter", "5"))


    if version == None:
        version = os.environ.get("version", "false").lower() == "true"

    if client_version == None:
        client_version = os.environ.get("client_version", "false").lower() == "true"

    if key_encyption == None:
        key_encyption = os.environ.get("key_encyption", "false").lower() == "true"



    if meta_datas == None:
        meta_datas = os.environ.get("meta_datas", "true").lower() == "true"

    if thread_number == None:
        thread_number = int(os.environ.get("thread_number", "1"))

    from upsonic import Upsonic_Remote
    return Upsonic_Remote(
        database_name, "https://cloud_3.upsonic.co", access_key, verify=True, locking=locking, client_id=client_id, cache=cache, cache_counter=cache_counter, version=version, client_version=client_version, key_encyption=key_encyption, meta_datas=meta_datas, quiet=quiet, thread_number=thread_number
    )  # pragma: no cover

def Upsonic_Cloud_Startup(database_name=None, access_key=None, locking=None, client_id=None, cache=None, cache_counter=5, version=None, client_version=None, key_encyption=None, meta_datas=None, quiet=False, thread_number=None):
    if database_name == None:
        database_name = os.environ.get("database_key")
    if access_key == None:
        access_key = os.environ.get("access_key")


    if locking == None:
        locking = os.environ.get("locking", "false").lower() == "true"


    if client_id == None:
        client_id = os.environ.get("client_id", None)

    if cache == None:
        cache = os.environ.get("cache", "true").lower() == "true"
    if cache_counter == None:
        cache_counter = int(os.environ.get("cache_counter", "5"))


    if version == None:
        version = os.environ.get("version", "false").lower() == "true"

    if client_version == None:
        client_version = os.environ.get("client_version", "false").lower() == "true"

    if key_encyption == None:
        key_encyption = os.environ.get("key_encyption", "false").lower() == "true"



    if meta_datas == None:
        meta_datas = os.environ.get("meta_datas", "true").lower() == "true"

    if thread_number == None:
        thread_number = int(os.environ.get("thread_number", "1"))

    from upsonic import Upsonic_Remote
    return Upsonic_Remote(
        database_name, "https://cloud_4.upsonic.co", access_key, verify=True, locking=locking, client_id=client_id, cache=cache, cache_counter=cache_counter, version=version, client_version=client_version, key_encyption=key_encyption, meta_datas=meta_datas, quiet=quiet, thread_number=thread_number
    )  # pragma: no cover




def Upsonic_Cloud_Readonly(database_name=None, access_key=None, locking=None, client_id=None, cache=None, cache_counter=5, version=None, client_version=None, key_encyption=None, meta_datas=None, quiet=False, thread_number=None):
    if database_name == None:
        database_name = os.environ.get("database_key")
    if access_key == None:
        access_key = os.environ.get("access_key")

    if locking == None:
        locking = os.environ.get("locking", "false").lower() == "true"


    if client_id == None:
        client_id = os.environ.get("client_id", None)

    if cache == None:
        cache = os.environ.get("cache", "true").lower() == "true"
    if cache_counter == None:
        cache_counter = int(os.environ.get("cache_counter", "5"))



    if version == None:
        version = os.environ.get("version", "false").lower() == "true"

    if client_version == None:
        client_version = os.environ.get("client_version", "false").lower() == "true"

    if key_encyption == None:
        key_encyption = os.environ.get("key_encyption", "false").lower() == "true"



    if meta_datas == None:
        meta_datas = os.environ.get("meta_datas", "true").lower() == "true"

    if thread_number == None:
        thread_number = int(os.environ.get("thread_number", "1"))

    from upsonic import Upsonic_Remote
    return Upsonic_Remote(
        database_name, "https://cloud_0.upsonic.co", access_key, verify=True, locking=locking, client_id=client_id, cache=cache, cache_counter=cache_counter, version=version, client_version=client_version, key_encyption=key_encyption, meta_datas=meta_datas, quiet=quiet, thread_number=thread_number
    )  # pragma: no cover




def Upsonic_Cloud_Generic(*args, **kwargs):
        _type = os.environ.get("type", "free")



        if _type.lower() == "free":

            
            return Upsonic_Cloud_Free(*args, **kwargs)
        elif _type.lower() == "pro":
            
            return Upsonic_Cloud_Pro(*args, **kwargs)

        elif _type.lower() == "premium":
            
            return Upsonic_Cloud_Premium(*args, **kwargs)

        elif _type.lower() == "startup":
            
            return Upsonic_Cloud_Startup(*args, **kwargs)

        elif _type.lower() == "readonly":
            
            return Upsonic_Cloud_Readonly(*args, **kwargs)

        else:
            raise Exception("Invalid type")




class _Upsonic_CLI:
    def __init__(self, type=None, database_name=None, access_key=None, locking=None, client_id=None) -> None:
        if type == None:
            self.type = os.environ.get("type", "free")


        if locking == None:
            self.locking = os.environ.get("locking", "false").lower() == "true"

        if client_id == None:
            self.client_id = os.environ.get("client_id", None)

        if database_name == None:
            database_name = os.environ.get("database_key")
        if access_key == None:
            access_key = os.environ.get("access_key")

        if self.type.lower() == "free":

            
            self.cloud = Upsonic_Cloud_Free(database_name, access_key, locking=self.locking, client_id=client_id)
        elif self.type.lower() == "pro":
            
            self.cloud = Upsonic_Cloud_Pro(database_name, access_key, locking=self.locking, client_id=client_id)

        elif self.type.lower() == "premium":
            
            self.cloud = Upsonic_Cloud_Premium(database_name, access_key, locking=self.locking, client_id=client_id)

        elif self.type.lower() == "startup":
            
            self.cloud = Upsonic_Cloud_Startup(database_name, access_key, locking=self.locking, client_id=client_id)

        elif self.type.lower() == "readonly":
            
            self.cloud = Upsonic_Cloud_Readonly(database_name, access_key, locking=self.locking, client_id=client_id)

        else:
            raise Exception("Invalid type")



    def lock(self, key):
        return self.cloud.lock_key(key)
        
    def unlock(self, key):
        return self.cloud.unlock_key(key)
               

    def set_set_version(self, version, client_id=None):
        self.cloud._log("Setting set version to " + str(version))
        return self.cloud.set_set_version(version, client_id=client_id)

    def set_get_version(self, version, client_id=None):
        self.cloud._log("Setting get version to " + str(version))
        if client_id == None:
            self.cloud._log("Client ID is None, setting global version")
        else:
            self.cloud._log("Client ID is " + str(client_id) + ", setting client version")
        return self.cloud.set_get_version(version, client_id=client_id)


    def print_set_version(self, client_id=None):
        return self.cloud.get_set_version_tag(client_id=client_id)

    def print_get_version(self, client_id=None):
        return self.cloud.get_get_version_tag(client_id=client_id)
    




    def update(self, module):
        backup = self.cloud.cache
        self.cloud.cache=False
        from upsonic_update import Upsonic_Update
        update = Upsonic_Update(self.cloud, pre_update_all=True, clear_olds=True) 

        import os
        import sys
        # add current directory to path
        top = os.getcwd()
        #Add top to sys.path top not last
        sys.path.insert(0, top)


        module_the = __import__(module)

        self.cloud.dump_module(module, module_the)

        update.update()
        self.cloud.cache=backup








def Upsonic_CLI():
    import fire
    fire.Fire(_Upsonic_CLI)


#!/usr/bin/python3
# -*- coding: utf-8 -*-

from setuptools import setup

with open('requirements.txt') as fp:
    install_requires = fp.read()
setup(
    name="upsonic",
    version="0.24.0",
    description="""Magic Cloud Layer""",
    long_description="".join(open("README.md", encoding="utf-8").readlines()),
    long_description_content_type="text/markdown",
    url="https://github.com/Upsonic/Upsonic",
    author="Upsonic",
    author_email="onur.atakan.ulusoy@upsonic.co",
    license="MIT",
    packages=["upsonic", "upsonic.remote", "upsonic.remote.localimport"],
    install_requires=install_requires,
    entry_points={
        "console_scripts": ["upsonic=upsonic.remote.interface:Upsonic_CLI"],
    },    
    python_requires=">= 3",
    zip_safe=False,
)







__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '1.7.6'

import copy
import glob
import os
import pkgutil
import sys
import traceback
import typing as t
import zipfile

if t.TYPE_CHECKING:
  from sys import _MetaPathFinder


def is_local(filename: str, pathlist: t.List[str]) -> bool:
  ''' Returns True if *filename* is a subpath of any of the paths in *pathlist*. '''

  filename = os.path.abspath(filename)
  for path_name in pathlist:
    path_name = os.path.abspath(path_name)
    if is_subpath(filename, path_name):
      return True
  return False


def is_subpath(path: str, parent: str) -> bool:
  ''' Returns True if *path* points to the same or a subpath of *parent*. '''

  try:
    relpath = os.path.relpath(path, parent)
  except ValueError:
    return False  # happens on Windows if drive letters don't match
  return relpath == os.curdir or not relpath.startswith(os.pardir)


def eval_pth(
  filename: str,
  sitedir: str,
  dest: t.Optional[t.List[str]] = None,
  imports: t.Optional[t.List[t.Tuple[str, int, str]]] = None,
) -> t.List[str]:
  ''' Evaluates a `.pth` file (including support for `import` statements), and appends the result to the list
  *dest*. If *dest* is #None, it will fall back to `sys.path`.

  If *imports* is specified, it must be a list. `import` statements will not executed but instead appended to
  that list in tuples of (*filename*, *line*, *stmt*).
  '''

  if dest is None:
    dest = sys.path

  if not os.path.isfile(filename):
    return []

  with open(filename, 'r') as fp:
    for index, line in enumerate(fp):
      if line.startswith('import'):
        if imports is None:
          exec_pth_import(filename, index+1, line)
        else:
          imports.append((filename, index+1, line))
      else:
        index = line.find('#')
        if index > 0: line = line[:index]
        line = line.strip()
        if not os.path.isabs(line):
          line = os.path.join(os.path.dirname(filename), line)
        line = os.path.normpath(line)
        if line and line not in dest:
          dest.insert(0, line)

  return dest


def exec_pth_import(filename: str, lineno: int, line: str) -> None:
  line = '\n' * (lineno - 1) + line.strip()
  try:
    exec(compile(line, filename, 'exec'))
  except BaseException:
    traceback.print_exc()


def extend_path(pth: t.List[str], name: str) -> t.List[str]:
  ''' Better implementation of #pkgutil.extend_path()  which adds support for zipped Python eggs. The original
  #pkgutil.extend_path() gets mocked by this function inside the #localimport context.
  '''

  def zip_isfile(z, name):
    name.rstrip('/')
    return name in z.namelist()

  pname = os.path.join(*name.split('.'))
  zname = '/'.join(name.split('.'))
  init_py = '__init__' + os.extsep + 'py'
  init_pyc = '__init__' + os.extsep + 'pyc'
  init_pyo = '__init__' + os.extsep + 'pyo'

  mod_path = list(pth)
  for path in sys.path:
    if zipfile.is_zipfile(path):
      try:
        egg = zipfile.ZipFile(path, 'r')
        addpath = (
          zip_isfile(egg, zname + '/__init__.py') or
          zip_isfile(egg, zname + '/__init__.pyc') or
          zip_isfile(egg, zname + '/__init__.pyo'))
        fpath = os.path.join(path, path, zname)
        if addpath and fpath not in mod_path:
          mod_path.append(fpath)
      except (zipfile.BadZipfile, zipfile.LargeZipFile):
        pass  # xxx: Show a warning at least?
    else:
      path = os.path.join(path, pname)
      if os.path.isdir(path) and path not in mod_path:
        addpath = (
          os.path.isfile(os.path.join(path, init_py)) or
          os.path.isfile(os.path.join(path, init_pyc)) or
          os.path.isfile(os.path.join(path, init_pyo)))
        if addpath and path not in mod_path:
          mod_path.append(path)

  return [os.path.normpath(x) for x in mod_path]


class localimport:

  def __init__(
    self,
    path: t.Union[t.List[str], str],
    parent_dir: t.Optional[str] = None,
    do_eggs: bool = True,
    do_pth: bool = True,
    do_autodisable: bool = True,
  ) -> None:
    if not parent_dir:
      frame = sys._getframe(1).f_globals
      if '__file__' in frame:
        parent_dir = os.path.dirname(os.path.abspath(frame['__file__']))

    # Convert relative paths to absolute paths with parent_dir and
    # evaluate .egg files in the specified directories.
    self.path = []
    if isinstance(path, str):
      path = [path]
    for path_name in path:
      if not os.path.isabs(path_name):
        if not parent_dir:
          raise ValueError('relative path but no parent_dir')
        path_name = os.path.join(parent_dir, path_name)
      path_name = os.path.normpath(path_name)
      self.path.append(path_name)
      if do_eggs:
        self.path.extend(glob.glob(os.path.join(path_name, '*.egg')))

    self.meta_path: t.List[_MetaPathFinder] = []
    self.modules: t.Dict[str, t.Any] = {}
    self.do_pth = do_pth
    self.in_context = False
    self.do_autodisable = do_autodisable
    self.pth_imports: t.List[t.Tuple[str, int, str]] = []

    if self.do_pth:
      seen = set()
      for path_name in self.path:
        for fn in glob.glob(os.path.join(path_name, '*.pth')):
          if fn in seen: continue
          seen.add(fn)
          eval_pth(fn, path_name, dest=self.path, imports=self.pth_imports)

  def __enter__(self) -> 'localimport':
    # pkg_resources comes with setuptools.
    try:
      import pkg_resources
      nsdict = copy.deepcopy(pkg_resources._namespace_packages)  # type: ignore
      declare_namespace = pkg_resources.declare_namespace
      pkg_resources.declare_namespace = self._declare_namespace  # type: ignore
    except ImportError:
      nsdict = None
      declare_namespace = None

    # Save the global importer state.
    self.state = {
      'nsdict': nsdict,
      'declare_namespace': declare_namespace,
      'nspaths': {},
      'path': sys.path[:],
      'meta_path': sys.meta_path[:],
      'disables': {},
      'pkgutil.extend_path': pkgutil.extend_path,
    }

    # Update the systems meta path and apply function mocks.
    sys.path[:] = self.path
    sys.meta_path[:] = self.meta_path + sys.meta_path
    pkgutil.extend_path = extend_path  # type: ignore

    # If this function is called not the first time, we need to
    # restore the modules that have been imported with it and
    # temporarily disable the ones that would be shadowed.
    for key, mod in list(self.modules.items()):
      try: self.state['disables'][key] = sys.modules.pop(key)
      except KeyError: pass
      sys.modules[key] = mod

    # Evaluate imports from the .pth files, if any.
    for fn, lineno, stmt in self.pth_imports:
      exec_pth_import(fn, lineno, stmt)

    # Add the original path to sys.path.
    sys.path += self.state['path']

    # Update the __path__ of all namespace modules.
    for key, mod in list(sys.modules.items()):
      if mod is None:
        # Relative imports could have lead to None-entries in
        # sys.modules. Get rid of them so they can be re-evaluated.
        prefix = key.rpartition('.')[0]
        if hasattr(sys.modules.get(prefix), '__path__'):
          del sys.modules[key]
      elif hasattr(mod, '__path__'):
        self.state['nspaths'][key] = copy.copy(mod.__path__)
        mod.__path__ = pkgutil.extend_path(mod.__path__, mod.__name__)

    self.in_context = True
    if self.do_autodisable:
      self.autodisable()
    return self

  def __exit__(self, *__) -> None:
    if not self.in_context:
      raise RuntimeError('context not entered')

    # Figure the difference of the original sys.path and the
    # current path. The list of paths will be used to determine
    # what modules are local and what not.
    local_paths = []
    for path in sys.path:
      if path not in self.state['path']:
        local_paths.append(path)
    for path in self.path:
      if path not in local_paths:
        local_paths.append(path)

    # Move all meta path objects to self.meta_path that have not
    # been there before and have not been in the list before.
    for meta in sys.meta_path:
      if meta is not self and meta not in self.state['meta_path']:
        if meta not in self.meta_path:
          self.meta_path.append(meta)

    # Move all modules that shadow modules of the original system
    # state or modules that are from any of the localimport context
    # paths away.
    modules = sys.modules.copy()
    for key, mod in modules.items():
      force_pop = False
      filename = getattr(mod, '__file__', None)
      if not filename and key not in sys.builtin_module_names:
        parent = key.rsplit('.', 1)[0]
        if parent in modules:
          filename = getattr(modules[parent], '__file__', None)
        else:
          force_pop = True
      if force_pop or (filename and is_local(filename, local_paths)):
        self.modules[key] = sys.modules.pop(key)

    # Restore the disabled modules.
    sys.modules.update(self.state['disables'])
    for key, mod in self.state['disables'].items():
      try: parent_name = key.split('.')[-2]
      except IndexError: parent_name = None
      if parent_name and parent_name in sys.modules:
        parent_module = sys.modules[parent_name]
        setattr(parent_module, key.split('.')[-1], mod)

    # Restore the original __path__ value of namespace packages.
    for key, path_list in self.state['nspaths'].items():
      try: sys.modules[key].__path__ = path_list
      except KeyError: pass

    # Restore the original state of the global importer.
    sys.path[:] = self.state['path']
    sys.meta_path[:] = self.state['meta_path']
    pkgutil.extend_path = self.state['pkgutil.extend_path']
    try:
      import pkg_resources
      pkg_resources.declare_namespace = self.state['declare_namespace']
      pkg_resources._namespace_packages.clear()  # type: ignore
      pkg_resources._namespace_packages.update(self.state['nsdict'])  # type: ignore
    except ImportError: pass

    self.in_context = False
    del self.state

  def _declare_namespace(self, package_name: str) -> None:
    '''
    Mock for #pkg_resources.declare_namespace() which calls
    #pkgutil.extend_path() afterwards as the original implementation doesn't
    seem to properly find all available namespace paths.
    '''

    self.state['declare_namespace'](package_name)
    mod = sys.modules[package_name]
    mod.__path__ = pkgutil.extend_path(mod.__path__, package_name)  # type: ignore

  def discover(self) -> t.Iterable[pkgutil.ModuleInfo]:
    return pkgutil.iter_modules(self.path)

  def disable(self, module: t.Union[t.List[str], str]) -> None:
    if not isinstance(module, str):
      for module_name in module:
        self.disable(module_name)
      return

    sub_prefix = module + '.'
    modules = {}
    for key, mod in sys.modules.items():
      if key == module or key.startswith(sub_prefix):
        try: parent_name = '.'.join(key.split('.')[:-1])
        except IndexError: parent_name = None

        # Delete the child module reference from the parent module.
        modules[key] = mod
        if parent_name and parent_name in sys.modules:
          parent = sys.modules[parent_name]
          try:
            delattr(parent, key.split('.')[-1])
          except AttributeError:
            pass

    # Pop all the modules we found from sys.modules
    for key, mod in modules.items():
      del sys.modules[key]
      self.state['disables'][key] = mod

  def autodisable(self) -> None:
    for loader, name, ispkg in self.discover():
      self.disable(name)


#!/usr/bin/python3
# -*- coding: utf-8 -*-

import json
import ast

from hashlib import sha256



import pickle
import os

import copy

import inspect
import pkgutil
import threading
import time
import textwrap

import cloudpickle


from contextlib import contextmanager

import sys

from rich.progress import Progress

class Upsonic_Remote:
    prevent_enable = False
    quiet_startup = False
    def _log(self, message):
        if not self.quiet:
            self.console.log(message)

    def __enter__(self):
        return self  # pragma: no cover

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass  # pragma: no cover

    def __init__(self, database_name, api_url, password=None, enable_hashing:bool=False, verify=True, locking=False, client_id=None, cache=False, cache_counter=None, version=False, client_version=False, key_encyption=False, meta_datas = True, quiet=False, thread_number=1):
        import requests
        from requests.auth import HTTPBasicAuth

        self.thread_number = thread_number

        self.quiet = quiet



        self.meta_datas = meta_datas

        self.force_compress = False
        self.force_encrypt = False
        self.key_encyption = key_encyption
        self.locking = locking
        self.enable_hashing = enable_hashing
        self.cache = cache
        self.cache_counter = cache_counter
        self._cache_counter = {}

        self.local_cache = {}
        
  
        self.cache_dir = os.path.join(os.getcwd(), "upsonic_cache")
        if not os.path.exists(self.cache_dir):
            os.mkdir(self.cache_dir)


        self.client_id = client_id

        self.verify = verify

        from upsonic import console

        self.console = console


        self.requests = requests
        self.HTTPBasicAuth = HTTPBasicAuth


        self.database_name = database_name
        if not Upsonic_Remote.quiet_startup:
            self._log(
                f"[{self.database_name[:5]}*] [bold white]Upsonic Cloud[bold white] initializing...",
            )
            
        if self.client_id is not None:
            if not Upsonic_Remote.quiet_startup:                
                self._log(f"[{self.database_name[:5]}*] [bold white]Client ID[bold white]: {self.client_id}")
        from upsonic import encrypt, decrypt
        self.encrypt = encrypt
        self.decrypt = decrypt


        self.api_url = api_url
        self.password = password

        try:
            self.informations = self._informations()
        except TypeError:
            self.informations = None
        if self.informations is not None:
            rate_limits = self.informations["rate_limit"][1]
            rate_limits = rate_limits.split(" ")
            self.rate_limit = int(rate_limits[1])
            self.thread_number = 10

        if not Upsonic_Remote.quiet_startup:
            self._log(
                f"[{self.database_name[:5]}*] [bold green]Upsonic Cloud[bold green] active",
            )
            self._log("---------------------------------------------")

        if self.cache:
            self.cache_hash_load()
            if self._cache_hash is None:
                self._cache_hash = {}
                self.cache_hash_save()



        self.version = version
        self.client_version = client_version


        self.enable_active = False




    def install_package(self, package):
        from pip._internal import main as pip

        package_name = package.split("==")[0]
        package_version = package.split("==")[1] if len(package.split("==")) > 1 else "Latest"

        the_dir = os.path.abspath(os.path.join(self.cache_dir, package_name, package_version))
        if not os.path.exists(the_dir):
            os.makedirs(the_dir)
        

        pip(["install", package, "--target", the_dir])






    @contextmanager
    def import_package(self, package):
        """
import sys
for a in list(sys.modules):
    if a.startswith("numpy"):
        del sys.modules[a]        
        """        


        package_name = package.split("==")[0]
        package_version = package.split("==")[1] if len(package.split("==")) > 1 else "Latest"

        the_dir = os.path.abspath(os.path.join(self.cache_dir, package_name, package_version))



        if not os.path.exists(the_dir):
            self.install_package(package)
        
        sys_path_backup = sys.path.copy()
        
        sys.path.insert(0, the_dir)

        try:
            yield
        finally:
            sys.path = sys_path_backup
    

    def extend_global(self, name, value):
        globals()[name] = value


    def load_module(self, module_name, encryption_key="a"):
        the_all = self.get_all(encryption_key=encryption_key)
        the_all_imports = {}
        for i, value in the_all.items():
            if "_upsonic_" in i:
                continue
            name = i.split(".")
            if module_name == name[0]:
                the_all_imports[i] = value
        import types
        def create_module_obj(dictionary):
            result = {}
            for key, value in dictionary.items():
                modules = key.split('.')
                current_dict = result
                for module in modules[:-1]:
                    if module not in current_dict:
                        current_dict[module] = types.ModuleType(module)
                    current_dict = vars(current_dict[module])
                current_dict[modules[-1]] = value
                

            return result


        generated_library = create_module_obj(the_all_imports)[module_name]


        return generated_library





    def enable(self, encryption_key="a", the_globals={}):
        the_keys = {}
        if Upsonic_Remote.prevent_enable:
            return the_keys.items()


        from upsonic import interface
        for key, value in the_globals.items():
            setattr(interface, key, value)
            globals()[key] = value


        self._log("[bold white]Syncing with Upsonic Cloud...")
        self.enable_active = True
        
        the_all_imports = {}


        the_us = self
        #Register the_us to globals
        globals()["the_us"] = the_us


        the_all = self.get_all(encryption_key=encryption_key)

        for key, value in the_all.items():
            if "_upsonic_" not in key:
                try:
                    original_key = key
                    original_key_without_dow = key.replace(".", "_")
                    
         
                    key = key.split(".")
                    message = value



                    from upsonic import interface
                    setattr(interface, key[-1], copy.copy(message))
                    the_keys[key[-1]] = copy.copy(message)


                except:
                    import traceback
                    traceback.print_exc()
                    self._log(f"[bold white]Error on patching '{key}'")

  
        return the_keys.items()


    def dump_module(self, module_name, module, encryption_key="a", compress=None, liberty=True):
        top_module = module



        cloudpickle.register_pickle_by_value(top_module)

        sub_modules = []
        if hasattr(top_module, "__path__"):

            for importer, modname, ispkg in pkgutil.walk_packages(path=top_module.__path__,
                                                        prefix=top_module.__name__+'.',
                                                        onerror=lambda x: None):
                sub_modules.append(importer.find_module(modname).load_module(modname))
        else:
            sub_modules.append(top_module)



        threads = []

        the_list=  []

        for sub_module in sub_modules:
            [the_list.append(obj) for name, obj in inspect.getmembers(sub_module)]

        
        # Extract just functions and classes
        the_list = [i for i in the_list if inspect.isfunction(i) or inspect.isclass(i)]
        # If the __module__ is not equal to module_name, remove it from the list
  
        the_list = [i for i in the_list if i.__module__.split(".")[0] == module_name]

        my_list = []
        for element in copy.copy(the_list):
                if inspect.isfunction(element):
                    name = element.__module__ +"." + element.__name__

                elif inspect.isclass(element):
                    name = element.__module__ +"." + element.__name__            
                if not "upsonic.remote" in name and not "upsonic_updater" in name and name != f"{module.__name__}.threading.Thread":
                    my_list.append(element)

        the_list = my_list



        with Progress() as progress:

            task1 = progress.add_task("           [red]Job Started...", total=len(the_list))
            task2 = progress.add_task("           [green]Job Complated...", total=len(the_list))




            for element in the_list:
                time.sleep(0.1)
                if inspect.isfunction(element):
                    name = element.__module__ +"." + element.__name__

                elif inspect.isclass(element):
                    name = element.__module__ +"." + element.__name__
                else:
                    continue
                
                first_element = name.split(".")[0]


    
                if first_element != module_name:
                    continue
                
                try:
                    while len(threads) >= self.thread_number:
                        for each in threads:
                            if not each.is_alive():
                                threads.remove(each)
                        time.sleep(0.1)
                
                    

                    the_thread = threading.Thread(target=self.set, args=(name, element), kwargs={"encryption_key": encryption_key, "compress": compress, "liberty": liberty})
                    the_thread.start()

                    thread = the_thread
                    threads.append(thread)
                    progress.update(task1, advance=1)

                except:
                    import traceback
                    traceback.print_exc()
                    self._log(f"[bold red]Error on '{name}'")
                    self.delete(name)
                    


            for each in threads:
                progress.update(task2, advance=1)
                each.join()


    
    def get_set_version_tag(self, client_id=None):
        the_key = "set_version_number"


        
        if client_id is not None:
            the_key = the_key + f"_{client_id}"

        else:
            if self.client_version:
                the_key = the_key + f"_{self.client_id}"            

        the_version = self.get(the_key, no_version=True)
        if the_version is None:
            return None
        if the_version == "latest":
            return None
        return the_version



    def get_get_version_tag(self, client_id=None):
        
        the_key = "get_version_number"


        
        if client_id is not None:
            the_key = the_key + f"_{client_id}"

        else:
            if self.client_version:
                the_key = the_key + f"_{self.client_id}"    


        the_version = self.get(the_key, no_version=True)
        if the_version is None:
            return None
        if the_version == "latest":
            return None
        return the_version


    def set_set_version(self, version_tag, client_id=None):
        the_key = "set_version_number"

        
        if client_id is not None:
            the_key = the_key + f"_{client_id}"

        else:
            if self.client_version:
                the_key = the_key + f"_{self.client_id}"    


        return self.set(the_key, version_tag, no_version=True)

    def set_get_version(self, version_tag, client_id=None):
        the_key = "get_version_number"

        
        if client_id is not None:
            the_key = the_key + f"_{client_id}"

        else:
            if self.client_version:
                the_key = the_key + f"_{self.client_id}"    


        return self.set(the_key, version_tag, no_version=True)


    def cache_hash_save(self):
        
        # Save the cache_hash to workdir/upsonic_cache_hash
        with open(os.path.join(self.cache_dir, "upsonic_cache_hash"), "wb") as f:
            pickle.dump(self._cache_hash, f)
    def cache_hash_load(self):
        # Load the cache_hash from workdir/upsonic_cache_hash
        try:
            with open(os.path.join(self.cache_dir, "upsonic_cache_hash"), "rb") as f:
                self._cache_hash = pickle.load(f)
        except FileNotFoundError:
            self._cache_hash = None
        

    def cache_set(self, key, value):
        self.local_cache[key] = value
        with open(f"{self.cache_dir}/{sha256(key.encode()).hexdigest()}", "wb") as f:
            pickle.dump(value, f)
    def cache_get(self, key):
        if key in self.local_cache:
            return self.local_cache[key]
        with open(f"{self.cache_dir}/{sha256(key.encode()).hexdigest()}", "rb") as f:
            return pickle.load(f)
    def cache_pop(self, key):
        if key in self.local_cache:
            self.local_cache.pop(key)
        try:
            os.remove(f"{self.cache_dir}/{sha256(key.encode()).hexdigest()}")
        except:
            pass



    def _informations(self):
        return self._send_request("GET", "/informations", make_json=True)

    def debug(self, message):
        data = {"message": message}
        return self._send_request("POST", "/controller/debug", data)

    def info(self, message):
        data = {"message": message}
        return self._send_request("POST", "/controller/info", data)

    def warning(self, message):
        data = {"message": message}
        return self._send_request("POST", "/controller/warning", data)

    def error(self, message):
        data = {"message": message}
        return self._send_request("POST", "/controller/error", data)

    def exception(self, message):
        data = {"message": message}
        return self._send_request("POST", "/controller/exception", data)

    def _send_request(self, method, endpoint, data=None, make_json=False):
        try:
            response = self.requests.request(
                method,
                self.api_url + endpoint,
                data=data,
                auth=self.HTTPBasicAuth("", self.password),
                verify=self.verify
            )

            try:
                response.raise_for_status()
                return response.text if not make_json else json.loads(response.text)
            except self.requests.exceptions.RequestException as e:  # pragma: no cover
                print(f"Error on '{self.api_url + endpoint}': ", response.text)
                return None  # pragma: no cover
        except self.requests.exceptions.ConnectionError:
            print("Error: Remote is down")
            return None


    def _lock_control(self, key, locking_operation=False):
        the_client_id = self.client_id
        if the_client_id is None:
            the_client_id = "Unknown"

        result = self.get(key+"_upsonic_lock", encryption_key=None, no_cache=True)
        if result is not None:
            result = result[1:]
            result = result[:-2]
            if result == the_client_id and not locking_operation:
                return False

            return True
        return False
  

    def lock_control(self, key):
        if self.locking:
            return self._lock_control(key)
        else:
            return False

    def lock_key(self, key):
        if self._lock_control(key, locking_operation=True):
            self._log(f"[bold red] '{key}' is already locked")
            return False

        the_client_id = self.client_id
        if the_client_id is None:
            the_client_id = "Unknown"

        if self.set(key+"_upsonic_lock", the_client_id, locking_operation=True, encryption_key=None, no_cache=True) == "Data set successfully":
            self._log(f"[bold green] '{key}' is locked")
            return True
        else:
            return False

    def unlock_key(self, key):
        result = self._lock_control(key, locking_operation=True)
        if not result:
            self._log(f"[bold red] '{key}' is already unlocked")
            return False
        
        if self._lock_control(key):
            self._log(f"[bold red] '{key}' is locked by another client")
            return False

    

        if self.delete(key+"_upsonic_lock") == "Data deleted successfully":
            self._log(f"[bold green] '{key}' is unlocked")
            return True
        else:         
            return False


    def _update_set(self, key, meta):
        if self.meta_datas:
            return self.set(key+"_upsonic_meta", meta, update_operation=True, encryption_key=None)


    def _liberty_set(self, key, liberty_class):
        if liberty_class:
            self.set(key+"_upsonic_liberty_class", True, update_operation=True, encryption_key=None)
        else:
            self.delete(key+"_upsonic_liberty_class")
        return self.set(key+"_upsonic_liberty", True, update_operation=True, encryption_key=None)


    def _liberty_unset(self, key, liberty_class):
        if liberty_class:
            self.delete(key+"_upsonic_liberty_class")
        self.delete(key+"_upsonic_liberty")




    def dump(self, key, value, encryption_key="a", compress=None, cache_policy=0, locking_operation=False, update_operation=False, version_tag=None, no_version=False, liberty=True):
        return self.set(key, value, encryption_key=encryption_key, compress=compress, cache_policy=cache_policy, locking_operation=locking_operation, update_operation=update_operation, version_tag=version_tag, no_version=no_version, liberty=liberty)


    def load(self, key, encryption_key="a", no_cache=False, version_tag=None, no_version=False):
        return self.get(key, encryption_key=encryption_key, no_cache=no_cache, version_tag=version_tag, no_version=no_version)


    def set(self, key, value, encryption_key="a", compress=None, cache_policy=0, locking_operation=False, update_operation=False, version_tag=None, no_version=False, liberty=True):
        self.cache_pop(key)
        if not locking_operation:
            if self.lock_control(key):
                self._log(f"[bold red] '{key}' is locked")
                return None

        the_type = type(value).__name__
        if the_type == "type":
            the_type = "class"

        meta = {'type_of_value': the_type}
        meta = json.dumps(meta)



        compress = True if self.force_compress else compress
        encryption_key = (
            self.force_encrypt if self.force_encrypt != False else encryption_key
        )

        
        liberty_class = inspect.isclass(value)
        if encryption_key is not None:
            value = self.encrypt(encryption_key, value, liberty=liberty)

        if not "_upsonic_" in key:
            if liberty:
                self._liberty_set(key, liberty_class)
            else:
                self._liberty_unset(key, liberty_class)


        key = sha256(key.encode()).hexdigest() if self.key_encyption else key

        data = {
            "database_name": self.database_name,
            "key": key,
            "value": value,
            "compress": compress,
            "cache_policy": cache_policy,
        }

        if version_tag is not None:
            copy_data = copy.copy(data)
            copy_data["key"] = copy_data["key"] + f"_upsonic_version_{version_tag}"
            self._send_request("POST", "/controller/set", copy_data)
            if not update_operation:
                self._update_set(copy_data["key"], meta)                   
        elif self.version and not no_version:
            the_version_ = self.get_set_version_tag()

            if the_version_ is not None:
                copy_data = copy.copy(data)
                copy_data["key"] = copy_data["key"] + f"_upsonic_version_{the_version_}"
 
                self._send_request("POST", "/controller/set", copy_data)   
                if not update_operation:
                    self._update_set(copy_data["key"], meta)                             

       
        if not update_operation:
            self._update_set(key, meta)
        return self._send_request("POST", "/controller/set", data)

    def get(self, key, encryption_key="a", no_cache=False, version_tag=None, no_version=False):
        




        key = sha256(key.encode()).hexdigest() if self.key_encyption else key

        if version_tag is not None:
            key = key + f"_upsonic_version_{version_tag}"
        elif self.version and not no_version:
            the_version_ = self.get_get_version_tag()
            if the_version_ is not None:
                key = key + f"_upsonic_version_{the_version_}"

        response = None
        if self.cache and not no_cache:
              
                if key not in self._cache_counter:
            
                    self._cache_counter[key] = 0
                self._cache_counter[key] = self._cache_counter[key] + 1
   
                if self._cache_counter[key] < self.cache_counter and self._cache_counter[key] != 1:
                    try:
          
                        response = self.cache_get(key)
          
                    except FileNotFoundError:
  
                        pass
                else:
                    if self._cache_counter[key] >= self.cache_counter:
                        self._cache_counter[key] = 1
                    the_hash = self.get(key+"_upsonic_updated", no_cache=True, no_version=True)
                    if key not in self._cache_hash:
                        self._cache_hash[key] = None
                    if the_hash != self._cache_hash[key] and the_hash is not None:
                        self._cache_hash[key] = the_hash
                        self.cache_hash_save()
                        self._log("Cache is updated")
                        try:
                            self.cache_pop(key)
                        except FileNotFoundError:
                            pass
                    else:
                        try:
                 
                            response = self.cache_get(key)
                 
                        except FileNotFoundError:
            
                            pass
                   

        encryption_key = (
            self.force_encrypt if self.force_encrypt != False else encryption_key
        )

        data = {"database_name": self.database_name, "key": key}

        if response is None:
    
            response = self._send_request("POST", "/controller/get", data)


        if self.cache:
            self.cache_set(key, response)

        if response is not None:
            if not response == "null\n":
                # Decrypt the received value
                if encryption_key is not None:
                    try:
                        response = self.decrypt(encryption_key, response)
                    except:
                        import traceback
                        traceback.print_exc()
                        pass                    
                return response
            else:
                return None

    def active(self, value=None, encryption_key="a", compress=None, just_name=False, liberty=True):
        def decorate(value):
            key = value.__name__ 
            if value.__module__ != "__main__" and value.__module__ != None and not just_name:
                key = value.__module__ + "." + key
            self.set(key, value, encryption_key=encryption_key, compress=compress, liberty=liberty)

        if value == None:
            return decorate
        else:
            decorate(value)
            return value

    def get_all(self, encryption_key="a"):
        encryption_key = (
            self.force_encrypt if self.force_encrypt != False else encryption_key
        )

        data = {"database_name": self.database_name}
        datas = self._send_request("POST", "/controller/get_all", data)

        datas = json.loads(datas)

        for each in datas:
            if encryption_key is not None:
                try:
                    datas[each] = self.decrypt(encryption_key, datas[each])
                except:
                    pass


        return datas

    def delete(self, key):
        data = {"database_name": self.database_name, "key": key}
        self.cache_pop(key)
        return self._send_request("POST", "/controller/delete", data)

    def database_list(self):
        return ast.literal_eval(self._send_request("GET", "/database/list"))


    def database_rename(self, database_name, new_database_name):
        data = {"database_name": database_name, "new_database_name": new_database_name}
        return self._send_request("POST", "/database/rename", data)


    def database_pop(self, database_name):
        data = {"database_name": database_name}
        return self._send_request("POST", "/database/pop", data)

    def database_pop_all(self):
        return self._send_request("GET", "/database/pop_all")

    def database_delete(self, database_name):
        data = {"database_name": database_name}
        return self._send_request("POST", "/database/delete", data)

    def database_delete_all(self):
        return self._send_request("GET", "/database/delete_all")


    
